use core::net::Ipv4Addr;

use dhcparse::{
    dhcpv4::{self, MessageType as DhcpMsgType},
    v4_options,
};
use embassy_net::{
    IpEndpoint, Stack,
    udp::{PacketMetadata, UdpMetadata, UdpSocket},
};
use embassy_time::Timer;
use heapless::Vec;
use log::info;
use pico_website::unwrap;

#[embassy_executor::task(pool_size = 1)]
pub async fn dhcp_server(stack: Stack<'static>) {
    let mut rx_buffer = [0; 4096];
    let mut tx_buffer = [0; 4096];
    let mut rx_meta = [PacketMetadata::EMPTY; 16];
    let mut tx_meta = [PacketMetadata::EMPTY; 16];
    let mut buf = [0; 4096];
    let mut res_buf = Vec::<u8, 4096>::new();
    loop {
        let mut socket = UdpSocket::new(
            stack,
            &mut rx_meta,
            &mut rx_buffer,
            &mut tx_meta,
            &mut tx_buffer,
        );
        unwrap(socket.bind(67)).await;

        loop {
            info!("receiving");
            let (n, _) = unwrap(socket.recv_from(&mut buf).await).await;
            info!("received");
            Timer::after_secs(0).await;

            let msg = unwrap(dhcpv4::Message::new(&buf[..n])).await;

            let msg_type = unwrap(v4_options!(msg; MessageType required)).await;
            for o in unwrap(msg.options()).await {
                info!(
                    "{:?}",
                    match o {
                        Err(_) => break,
                        Ok(o) => o,
                    }
                );
            }
            info!("{:?}", msg_type);
            Timer::after_secs(0).await;

            match msg_type {
                DhcpMsgType::DISCOVER | DhcpMsgType::REQUEST => {
                    res_buf.clear();
                    res_buf.push(2).unwrap(); // op
                    res_buf.push(1).unwrap(); // htype
                    res_buf.push(6).unwrap(); // hlen
                    res_buf.push(0).unwrap(); // hops
                    res_buf.extend_from_slice(&buf[4..8]).unwrap(); // xid
                    res_buf.extend_from_slice(&[0; 2]).unwrap(); // secs
                    res_buf.extend_from_slice(&[0x80, 0x00]).unwrap(); // flags
                    res_buf.extend_from_slice(&buf[12..16]).unwrap(); // ciaddr
                    res_buf.extend_from_slice(&[192, 254, 0, 12]).unwrap(); // yiaddr
                    res_buf.extend_from_slice(&[0; 4]).unwrap(); // siaddr
                    res_buf.extend_from_slice(&buf[24..28]).unwrap(); // giaddr
                    res_buf.extend_from_slice(&buf[28..44]).unwrap(); // chaddr
                    res_buf.extend_from_slice(&[0; 192]).unwrap(); // sname/file
                    res_buf.extend_from_slice(&[99, 130, 83, 99]).unwrap(); // magic cookie

                    res_buf
                        .extend_from_slice(&[
                            53,
                            1,
                            match msg_type {
                                DhcpMsgType::DISCOVER => 2, // DHCP_OFFER
                                DhcpMsgType::REQUEST => 5,  // DHCP_ACK
                                _ => unreachable!(),
                            },
                        ])
                        .unwrap(); // opt message type
                    unwrap(
                        write_dhcp_opts(
                            &mut res_buf,
                            match msg_type {
                                DhcpMsgType::DISCOVER => {
                                    unwrap(v4_options!(msg; ParameterRequestList))
                                        .await
                                        .unwrap_or(&[])
                                }
                                DhcpMsgType::REQUEST => &[1, 3, 51, 6, 54],
                                _ => unreachable!(),
                            },
                        )
                        .await,
                    )
                    .await;
                    res_buf.push(255).unwrap(); // end option

                    for o in unwrap(unwrap(dhcpv4::Message::new(&res_buf)).await.options()).await {
                        info!(
                            "{:?}",
                            match o {
                                Err(_) => break,
                                Ok(o) => o,
                            }
                        );
                        Timer::after_secs(0).await;
                    }

                    unwrap(
                        socket
                            .send_to(
                                &res_buf,
                                UdpMetadata {
                                    endpoint: IpEndpoint::new(
                                        Ipv4Addr::new(255, 255, 255, 255).into(),
                                        68,
                                    ),
                                    local_address: Some(Ipv4Addr::new(192, 254, 0, 2).into()),
                                    meta: Default::default(),
                                },
                            )
                            .await,
                    )
                    .await;
                    info!("offer sent");
                    Timer::after_secs(0).await;
                }
                _ => {}
            }

            // unwrap(socket.send_to(&buf[..n], ep).await).await;
        }
    }
}

async fn write_dhcp_opts<const N: usize>(buf: &mut Vec<u8, N>, op_codes: &[u8]) -> Result<(), ()> {
    for o in op_codes {
        let (opt_len, opt): (u8, &[u8]) = match o {
            1 => (4, &[255, 255, 255, 0]), // DhcpOption::SubnetMask(&dhcpv4::Addr([255, 255, 255, 0])),
            2 => (4, &3600_i32.to_be_bytes()), // DhcpOption::TimeOffset(3600),
            3 => (4, &[192, 254, 0, 2]),   // DhcpOption::Router(&[dhcpv4::Addr([192, 254, 0, 2])]),
            6 => (4, &[0, 0, 0, 0]), // DhcpOption::DomainNameServer(&[dhcpv4::Addr([0, 0, 0, 0])]),
            12 => (4, b"blue"),      // DhcpOption::HostName(b"blue"),
            15 => (11, b"LocalDomain"), // DhcpOption::DomainName(b"LocalDomain"),
            26 => (2, &1514_u16.to_be_bytes()), // DhcpOption::Unknown(26, &[0x5, 0xEA]), // mtu
            28 => (4, &[192, 254, 0, 255]), // DhcpOption::Unknown(28, &[192, 254, 0, 255]), // broadcast
            51 => (4, &700_u32.to_be_bytes()), // DhcpOption::AddressLeaseTime(700),
            54 => (4, &[192, 254, 0, 2]), // DhcpOption::ServerIdentifier(&dhcpv4::Addr([192, 254, 0, 2])),
            58 => (4, &500_u32.to_be_bytes()), // DhcpOption::Unknown(58, &[0, 0, 0x1, 0xF4]), // renewal time = 500s
            59 => (4, &600_u32.to_be_bytes()), // DhcpOption::Unknown(59, &[0, 0, 0x2, 0x58]), // rebinding time = 600s
            _ => {
                info!("Unhandled requested option : {}", o);
                Timer::after_secs(0).await;
                continue;
            }
        };
        buf.push(*o).map_err(|_| ())?;
        buf.push(opt_len).map_err(|_| ())?;
        buf.extend_from_slice(opt)?;
    }
    Ok(())
}
